/*
    Copyright (C) 2002, Steve Lhomme <steve.lhomme @ free.fr>
    This file is part of mpa2wav.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
    \file processus.cpp
    Convert MPEG I/II Layer 1/2/3 to RIFF/WAV/MPEG files
    \todo remove bad frames
    \todo the ID3v1 and v2 data could be put in a special WAV chunk
*/
#include <stdio.h>
#include <cassert>

#include "processus.h"
#include "mpafile.h"
#include "mpiframe.h"
#include "mpaframe.h"
#include "mpg_wav.h"

#if !defined(NO_NAMESPACE)
using namespace std;
using namespace M2W_NAMESPACE;
#endif

const uint32 BUFFER_SIZE = 5000;

#ifdef  __cplusplus
extern "C" {
#endif

int process(const string & input_Filename, const string & output_Filename, bool bKeepAll = false, bool bVerbose = true, bool bVerboseMore = true)
{
    if (bVerbose)
	printf("Converting %s to %s\r\n", input_Filename.c_str(), output_Filename.c_str());

    // open the input file
    MPA_File input(input_Filename, mode_read);

    // open the output file
    MPG_WAV_file output(output_Filename);
    if (!output.Open(FILE_MODE_WRITE | FILE_MODE_ERASE))
    {
	printf("Impossible to open output file '%s'\r\n", output_Filename.c_str());
	return -1;
    }

    // prepare to read the whole file for statistics
    if (bVerbose)
	printf("Reading statistics in %s\r\n", input_Filename.c_str());

    input.SkipToFrameStart();

    bool bInfoFrameOn = false;
    MPA_Lame_Frame infoFrame;
    if (input.CheckInfoFrame())
    {
	*static_cast<MPA_Frame *>(&infoFrame) = input.GetInfoFrame();
	bInfoFrameOn = true;
    }

    uint32 FrameNumber = 0;
    uint64 SampleNumbers = 0;
    uint64 TotalFrameSize = 0;
    uint16 SampleSize = 0;
    MPA_Frame frame;

    while (input.FindNextSyncWord())
    {
	FrameNumber++;
	frame = input.GetFrame();
	if (input.Feed())
	{
	    /// \todo handle the bitrate of the frame in a key/value table, so that we can detect real CBR files
	    TotalFrameSize += frame.DataSize() + 4;
	    SampleNumbers += frame.SampleSize();
	    SampleSize += frame.SampleSize();
	}

	input.SkipData();
    }
    input.FixStats();

    uint64 totalByteSize = TotalFrameSize * input.SamplingRate();

    if (bVerboseMore)
    {
	printf("Sampling Rate = %d / Total Valid Frame Size = %ld\r\n",
	       input.SamplingRate(), TotalFrameSize);
	printf("Average Bitrate = %d (8 x %d)\r\n", input.AverageBitRate(), uint32(0.5 + (input.AverageBitRate() / 8)));
	printf("Sample number = %d\r\n", SampleNumbers);
	printf("FrameNumber = %d\r\n", FrameNumber);
    }

    // construct the WAV header with the statistics

    // prepare to read for output
    input.SkipToFrameStart();

    /// \todo add the option to skip the Info frame or not
    if (bInfoFrameOn)
    {
	input.FindNextSyncWord();
	input.GetFrame();
	input.SkipData();
    }

    if (bVerbose)
	printf("Writing frames in %s\r\n", output_Filename.c_str());

    /// \todo move this somewhere else (cleaner)
    output.setFrameType(input.FrameType());
    output.setByterate(0.5 + (input.AverageBitRate() / 8));
    output.setSampleSize(SampleSize);

    output.WriteHead(TotalFrameSize, SampleNumbers, false);

    binary frame_buffer[BUFFER_SIZE]; // should be enough for 32kHz/320kbps
    if (!bKeepAll)
    {
	while (input.FindNextSyncWord())
	{
	    frame = input.GetFrame();
	    if (input.Test())
	    {
		output.Write(frame, 4);
		assert(frame.DataSize() < BUFFER_SIZE);
		input.GetData(frame_buffer, BUFFER_SIZE);
		output.Write(frame_buffer, frame.DataSize());
	    }
	    else input.SkipData();
	}
    }
    else
    {
	uint32 sizeRead;
	do
	{
	    sizeRead = input.read(frame_buffer, BUFFER_SIZE);
	    output.Write(frame_buffer, sizeRead);
	} while (sizeRead == BUFFER_SIZE);
    }

    output.Close();

    return 0;
}

#ifdef  __cplusplus
}
#endif
